# SPDX-License-Identifier: GPL-3.0-or-later
"""
Package containing all the builders.

isort:skip_file
"""

# pylint: disable=unused-import
from . import a20_olinuxino_lime2
from . import a20_olinuxino_lime
from . import a20_olinuxino_micro
from . import amd64
from . import arm
from . import arm64
from . import armhf
from . import banana_pro
from . import beaglebone
from . import cubieboard2
from . import cubietruck
from . import i386
from . import lamobo_r1
from . import orange_pi_zero
from . import pcduino3
from . import pine64_plus
from . import pine64_lts
from . import qemu_amd64
from . import qemu_i386
from . import raspberry_pi_2
from . import raspberry_pi_3
from . import raspberry_pi_3_b_plus
from . import raspberry_pi_with_uboot
from . import vagrant
from . import virtualbox_amd64
from . import virtualbox_i386



#!/usr/bin/python3
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Tool to build FreedomBox images for various targets.
"""

from .application import Application

__version__ = '0.31'

__all__ = [
    'Application',
    'main',
]


def main():
    Application().run()



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Tests for miscellaneous utility methods.
"""

import unittest

from freedommaker import utils


class TestUtils(unittest.TestCase):
    """Test for miscellaneous utility methods."""
    def test_add_disk_offsets(self):
        """Test disk offset addition."""
        self.assertEqual(utils.add_disk_offsets('1MiB', '2MiB'), '3MiB')
        # mib is not allowed as unit. Only MiB.
        self.assertRaises(NotImplementedError, utils.add_disk_offsets, '1mib',
                          '1mib')
        self.assertRaises(NotImplementedError, utils.add_disk_offsets, '1GiB',
                          '1MiB')
        self.assertRaises(NotImplementedError, utils.add_disk_offsets, '1MiB',
                          '1GiB')
        self.assertRaises(ValueError, utils.add_disk_offsets, 'xMiB', '1MiB')
        self.assertRaises(ValueError, utils.add_disk_offsets, 'xMiB', 'xMiB')

    def test_parse_disk_size(self):
        """Test parsing disk sizes."""
        self.assertEqual(utils.parse_disk_size(0), 0)
        self.assertEqual(utils.parse_disk_size(123), 123)
        self.assertEqual(utils.parse_disk_size('123'), 123)
        self.assertEqual(utils.parse_disk_size('123b'), 123)
        self.assertEqual(utils.parse_disk_size('123k'), 123 * 1024)
        self.assertEqual(utils.parse_disk_size('123K'), 123 * 1024)
        self.assertEqual(utils.parse_disk_size('123M'), 123 * 1024 * 1024)
        self.assertEqual(utils.parse_disk_size('123G'),
                         123 * 1024 * 1024 * 1024)
        self.assertEqual(utils.parse_disk_size('123T'),
                         123 * 1024 * 1024 * 1024 * 1024)
        self.assertRaises(ValueError, utils.parse_disk_size, '-1M')
        self.assertRaises(ValueError, utils.parse_disk_size, '1B')
        self.assertRaises(ValueError, utils.parse_disk_size, '1P')
        self.assertRaises(ValueError, utils.parse_disk_size, 'test')

    def test_format_disk_size(self):
        """Test that formatting disk sizes works."""
        self.assertEqual(utils.format_disk_size(0), '0')
        self.assertEqual(utils.format_disk_size(1), '1')
        self.assertEqual(utils.format_disk_size(1023), '1023')
        self.assertEqual(utils.format_disk_size(1024), '1K')
        self.assertEqual(utils.format_disk_size(1023 * 1023), '1046529')
        self.assertEqual(utils.format_disk_size(1023 * 1024), '1023K')
        self.assertEqual(utils.format_disk_size(1024 * 1024), '1M')
        self.assertEqual(utils.format_disk_size(1023 * 1024 * 1024), '1023M')
        self.assertEqual(utils.format_disk_size(1024 * 1024 * 1024), '1G')
        self.assertEqual(utils.format_disk_size(1023 * 1024 * 1024 * 1024),
                         '1023G')
        self.assertEqual(utils.format_disk_size(1024 * 1024 * 1024 * 1024),
                         '1T')
        self.assertEqual(
            utils.format_disk_size(1023 * 1024 * 1024 * 1024 * 1024), '1023T')

    def test_add_disk_sizes(self):
        """Test that adding two disk sizes works as expected."""
        self.assertEqual(utils.add_disk_sizes('1M', '1K'), '1025K')
        self.assertEqual(utils.add_disk_sizes('1G', '1K'), '1048577K')
        self.assertEqual(utils.add_disk_sizes('512M', '512M'), '1G')
        self.assertEqual(utils.add_disk_sizes('3800M', '1000M'), '4800M')



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Library of small actions that make up image building.

Most simple wrappers over command line utilities without any understanding of
state of build process.
"""

import contextlib
import logging
import os
import re
import shutil
import tempfile

import cliapp

logger = logging.getLogger(__name__)

STABLE_CODENAME = 'bullseye'


def run(*args, **kwargs):
    """Run a command."""
    def log_stdout(data):
        """Handle stdout data from executed command."""
        logger.debug('> %s', str(data.decode().strip()))

    def log_stderr(data):
        """Handle stderr data from executed command."""
        logger.debug('2> %s', str(data.decode().strip()))

    logger.info('Executing command - %s %s', args, kwargs)
    environ = kwargs['env'] if 'env' in kwargs else os.environ.copy()
    environ['LC_ALL'] = 'C'
    environ['LANGUAGE'] = 'C'
    environ['LANG'] = 'C'
    environ['DEBIAN_FRONTEND'] = 'noninteractive'
    environ['DEBCONF_NONINTERACTIVE_SEEN'] = 'true'
    kwargs['env'] = environ
    kwargs['stdout_callback'] = log_stdout
    kwargs['stderr_callback'] = log_stderr
    return cliapp.runcmd(*args, **kwargs)


def run_in_chroot(state, *args, **kwargs):
    """Run a command inside chroot of mount point."""
    args = [['chroot', state['mount_point']] + arg for arg in args]
    return run(*args, **kwargs)


def path_in_mount(state, path):
    """Return the path inside the mount point of image.

    If an absolute path is provided as path, it will returned as it is.

    """
    return os.path.join(state['mount_point'], path)


def schedule_cleanup(state, method, *args, **kwargs):
    """Make a note of the cleanup operations to happen."""
    state.setdefault('cleanup', []).append([method, args, kwargs])


def cleanup(state):
    """Run all the scheduled cleanups in reverse order."""
    state.setdefault('cleanup', [])
    for cleanup_step in reversed(state['cleanup']):
        method, args, kwargs = cleanup_step
        method(*args, **kwargs)


def create_ram_directory_image(state, image_file, size):
    """Create a temporary RAM directory."""
    logger.info('Create RAM directory for image: %s (%s)', image_file, size)
    directory = tempfile.TemporaryDirectory()
    run([
        'mount', '-o', 'size=' + size, '-t', 'tmpfs', 'tmpfs', directory.name
    ])

    state['ram_directory'] = directory
    temp_image_file = os.path.join(directory.name,
                                   os.path.basename(image_file))
    state['image_file'] = temp_image_file

    schedule_cleanup(state, remove_ram_directory, directory)
    schedule_cleanup(state, copy_image, state, temp_image_file, image_file)


def remove_ram_directory(directory):
    """Remove RAM directory created for temporary image path."""
    logger.info('Cleanup RAM directory %s', directory)
    run(['umount', directory.name])
    directory.cleanup()


def copy_image(state, source_image, target_image):
    """Copy from temp image to target path."""
    if not state['success']:
        target_image += '.failed'

    logger.info('Copying file: %s -> %s', source_image, target_image)
    run(['cp', '--sparse=always', source_image, target_image])


def create_temp_image(state, image_file):
    """Create a temp image for a target image file on disk."""
    temp_image_file = image_file + '.temp'
    logger.info('Creating temp image %s for image %s', temp_image_file,
                image_file)
    state['image_file'] = temp_image_file

    schedule_cleanup(state, move_image, state, temp_image_file, image_file)


def move_image(state, source_image, target_image):
    """Move from temp image name to final image name."""
    if not state['success']:
        target_image += '.failed'

    logger.info('Moving image: %s -> %s', source_image, target_image)
    run(['mv', source_image, target_image])


def create_image(state, size):
    """Create an empty sparse file using qemu-image."""
    logger.info('Creating image %s of size %s', state['image_file'], size)
    run(['qemu-img', 'create', '-f', 'raw', state['image_file'], size])


def create_partition_table(state, partition_table_type):
    """Create an empty partition table in given device."""
    logger.info('Creating partition table on %s of type %s',
                state['image_file'], partition_table_type)
    run(['parted', '-s', state['image_file'], 'mklabel', partition_table_type])


def create_partition(state, label, start, end, filesystem_type):
    """Create a primary partition in a given device."""
    filesystem_map = {'vfat': 'fat32'}
    filesystem_type = filesystem_map.get(filesystem_type, filesystem_type)

    partition_type = 'primary'
    logger.info('Creating partition %s in %s (range %s - %s) of type %s',
                label, state['image_file'], start, end, filesystem_type)
    run([
        'parted', '-s', state['image_file'], 'mkpart', partition_type,
        filesystem_type, start, end
    ])

    state.setdefault('partitions', []).append(label)


def set_boot_flag(state, partition_number):
    """Set boot flag on a partition of a device."""
    logger.info('Setting boot flag on %s partition for %s', partition_number,
                state['image_file'])
    run([
        'parted', '-s', state['image_file'], 'set',
        str(partition_number), 'boot', 'on'
    ])


def loopback_setup(state):
    """Perform mapping to loopback devices from partitions in image file."""
    logger.info('Setting up loopback mappings for %s', state['image_file'])
    output = run(['kpartx', '-asv', state['image_file']]).decode()
    loop_device = None
    devices = []
    partition_number = 0
    for line in output.splitlines():
        columns = line.split()
        if columns[0] == 'add':
            device = '/dev/mapper/{}'.format(columns[2])
            label = state['partitions'][partition_number]
            state.setdefault('devices', {})[label] = device
            devices.append(device)
            partition_number += 1
            if not loop_device:
                loop_device = re.match(r'^(loop\d+)p\d+$', columns[2])[1]
                loop_device = '/dev/{}'.format(loop_device)
                state['loop_device'] = loop_device

    # Cleanup runs in reverse order
    if loop_device:
        schedule_cleanup(state, force_release_loop_device, loop_device)

    for device in devices:
        schedule_cleanup(state, force_release_partition_loop, device)

    schedule_cleanup(state, loopback_teardown, state['image_file'])


def force_release_partition_loop(loop_device):
    """Force release a partition mapping on a loop device."""
    logger.info('Force releasing partition on loop device %s', loop_device)
    run(['dmsetup', 'remove', loop_device], ignore_fail=True)


def force_release_loop_device(loop_device):
    """Force release of a loop setup for entire device"""
    logger.info('Force releasing loop setup for device %s', loop_device)
    run(['losetup', '-d', loop_device], ignore_fail=True)


def loopback_teardown(image_file):
    """Unmap loopback devices from partitions in image file."""
    logger.info('Tearing down loopback mappings for %s', image_file)
    run(['kpartx', '-dsv', image_file])


def create_filesystem(device, filesystem_type):
    """Create a filesystem on a given device."""
    logger.info('Creating filesystem on %s of type %s', device,
                filesystem_type)
    run(['mkfs', '-t', filesystem_type, device])

    # Due to a bug, probably in udev, /dev/disk/by-uuid/<uuid> link is not
    # reliably created after the creation of the filesystem. This leads to
    # update-grub using root=/dev/mapper/loop0p1 instead of root=UUID=<uuid>
    # when creating grub.cfg. This results in an unbootable image. Force udev
    # events as a workaround.
    run(['udevadm', 'trigger', device])
    run(['udevadm', 'settle'])


def mount_filesystem(state,
                     label_or_path,
                     sub_mount_point,
                     is_bind_mount=False):
    """Mount a device on a mount point."""
    if not sub_mount_point:
        mount_point = tempfile.mkdtemp()
        state['mount_point'] = mount_point
    else:
        mount_point = path_in_mount(state, sub_mount_point)
        os.makedirs(mount_point, exist_ok=True)

    if not is_bind_mount:
        device = state['devices'][label_or_path]
        options = []
    else:
        device = label_or_path
        options = ['-o', 'bind']

    logger.info('Mounting device %s on %s with options %s', device,
                mount_point, options)
    run(['mount', device, mount_point] + options)
    state.setdefault('sub_mount_points', {})[label_or_path] = sub_mount_point

    schedule_cleanup(state, unmount_filesystem, device, mount_point)


def unmount_filesystem(device, mount_point, ignore_fail=False):
    """Unmount a filesystem."""
    logger.info('Unmounting device %s from mount point %s', device,
                mount_point)
    run(['umount', mount_point], ignore_fail=ignore_fail)


def process_cleanup(state):
    """Kill all processes using a given mount point."""
    mount_point = state['mount_point']
    logger.info('Killing all processes on the mount point %s', mount_point)
    run(['fuser', '-mvk', mount_point], ignore_fail=True)
    # XXX: Twice seems to work better?
    run(['fuser', '-mvk', mount_point], ignore_fail=True)


def setup_extra_storage(state, file_system_type, size):
    """Add an extra storage device to a btrfs filesystem."""
    if file_system_type != 'btrfs':
        return

    extra_storage_file = state['image_file'] + '.extra'
    mount_point = state['mount_point']
    logger.info('Adding extra storage to file system %s', mount_point)
    run(['qemu-img', 'create', '-f', 'raw', extra_storage_file, size])
    output = run(['losetup', '--show', '--find', extra_storage_file])
    loop_device = output.decode().strip()
    run(['btrfs', 'device', 'add', loop_device, mount_point])

    schedule_cleanup(state, cleanup_extra_storage, state, loop_device,
                     extra_storage_file)


def cleanup_extra_storage(state, loop_device, extra_storage_file):
    """Remove the extra storage added to a btrfs filesystem and balance it."""
    mount_point = state['mount_point']
    logger.info('Removing extra storage from file system %s', mount_point)

    _btrfs_rebalance(mount_point)
    run(['btrfs', 'balance', 'start', '-mconvert=dup', mount_point],
        ignore_fail=True)

    # Remove the extra storage device from btrfs filesystem
    run(['btrfs', 'device', 'remove', loop_device, mount_point])
    run(['losetup', '--detach', loop_device])
    run(['rm', '-f', extra_storage_file])

    _btrfs_rebalance(mount_point)


def _btrfs_rebalance(mount_point):
    """Re-balance btrfs filesystem."""
    for usage in range(0, 100, 20):
        run(['btrfs', 'balance', 'start', f'-musage={usage}', mount_point],
            ignore_fail=True)
        run(['btrfs', 'balance', 'start', f'-dusage={usage}', mount_point],
            ignore_fail=True)


def debootstrap(state, architecture, distribution, variant, components,
                packages, mirror):
    """Debootstrap into a mounted directory."""
    target = state['mount_point']
    logger.info(
        'Debootstraping into %s, architecture %s, '
        'distribution %s, variant %s, components %s, build mirror %s', target,
        architecture, distribution, variant, components, mirror)
    try:
        run([
            'debootstrap', '--arch=' + architecture, '--variant=' + variant,
            '--components=' + ','.join(components),
            '--include=' + ','.join(packages), distribution, target, mirror
        ])
    except (Exception, KeyboardInterrupt):
        logger.info(
            'Unmounting filesystems that may have been left by debootstrap')
        for path in ('proc', 'sys', 'etc/machine-id'):
            unmount_filesystem(None,
                               os.path.join(target, path),
                               ignore_fail=True)
        raise

    # During bootstrap, /etc/machine-id path might be bind mounted.
    schedule_cleanup(state,
                     unmount_filesystem,
                     None,
                     os.path.join(target, 'etc/machine-id'),
                     ignore_fail=True)


@contextlib.contextmanager
def no_run_daemon_policy(state):
    """Context manager to ensure daemons are not run during installs."""
    logger.info('Enforcing policy not to run daemons')
    path = path_in_mount(state, 'usr/sbin/policy-rc.d')
    content = '''#!/bin/sh
exit 101
'''
    with open(path, 'w') as file_handle:
        file_handle.write(content)

    os.chmod(path, 0o755)
    yield

    logger.info('Relaxing policy not to run daemons')
    os.unlink(path)


def install_from_backports(state, packages):
    """Install a package using apt from backports."""

    with no_run_daemon_policy(state):
        try:
            logger.info('Installing packages from backports: %s',
                        " ".join(packages))
            run_in_chroot(state, [
                'apt-get', 'install', '-y', *[
                    f'{package}/{STABLE_CODENAME}-backports'
                    for package in packages
                ]
            ])
        except cliapp.AppException as ex:
            logger.info(
                'Packages not found in backports. Installing from stable.')
            run_in_chroot(state, ['apt-get', 'install', '-y', *packages])

    run_in_chroot(state, ['apt-get', 'autoremove', '-y'])


def install_package(state, package):
    """Install a package using apt."""
    logger.info('Installing package %s', package)

    with no_run_daemon_policy(state):
        run_in_chroot(state, ['apt-get', 'install', '-y', package])


def install_custom_package(state, package_path):
    """Install a custom .deb file."""
    logger.info('Install custom .deb package %s', package_path)
    sub_destination = os.path.join('tmp', os.path.basename(package_path))
    destination_path = path_in_mount(state, sub_destination)
    shutil.copyfile(package_path, destination_path)
    package_path = os.path.join('/tmp', os.path.basename(package_path))
    with no_run_daemon_policy(state):
        run_in_chroot(
            state,
            ['apt', 'install', '-y', '--allow-downgrades', package_path])


def set_hostname(state, hostname):
    """Set the hostname inside the image."""
    logger.info('Setting hostname to %s', hostname)

    etc_hostname_path = path_in_mount(state, 'etc/hostname')
    with open(etc_hostname_path, 'w') as file_handle:
        file_handle.write(hostname + '\n')

    etc_hosts_path = path_in_mount(state, 'etc/hosts')
    with open(etc_hosts_path, 'r') as file_handle:
        lines = file_handle.readlines()
    with open(etc_hosts_path, 'w') as file_handle:
        appended = False
        for line in lines:
            if line.startswith('127.0.1.1'):
                line = line + ' ' + hostname
                appended = True

            file_handle.write(line)

        if not appended:
            file_handle.write('127.0.1.1 ' + hostname + '\n')


def get_fstab_options(filesystem_type):
    """Return options to use for a filesystem type."""
    if filesystem_type == 'btrfs':
        # Enable btrfs transparent compression. See:
        # https://fedoraproject.org/wiki/Changes/BtrfsTransparentCompression
        flags = ['compress=zstd:1']
    else:
        flags = ['errors=remount-ro']

    return ','.join(flags) or 'defaults'


def get_uuid_of_device(device):
    """Return the UUID of a given device."""
    output = run(['blkid', '--output=value', '--match-tag=UUID', device])
    return output.decode().strip()


def add_fstab_entry(state, label, filesystem_type, pass_number, append=True):
    """Add an entry in /etc/fstab for a disk partition."""
    file_path = path_in_mount(state, 'etc/fstab')
    device = 'UUID={}'.format(get_uuid_of_device(state['devices'][label]))
    values = {
        'device': device,
        'mount_point': '/' + (state['sub_mount_points'][label] or ''),
        'filesystem_type': filesystem_type,
        'options': get_fstab_options(filesystem_type),
        'frequency': '0',
        'pass_number': pass_number
    }
    line = '{device} {mount_point} {filesystem_type} {options} {frequency} ' \
           '{pass_number}\n'
    line = line.format(**values)
    logger.info('Adding fstab entry %s', values)

    mode = 'a' if append else 'w'
    with open(file_path, mode) as file_handle:
        file_handle.write(line)


def install_grub(state, target=None, is_efi=False):
    """Install grub boot loader on the loop back device."""
    device = state['loop_device']
    logger.info('Installing grub boot loader on device %s', device)
    run_in_chroot(state, ['update-grub'])
    args = [f'--target={target}'] if target else []
    args += ['--no-nvram'] if is_efi else []
    run_in_chroot(state, ['grub-install', device] + args)


def setup_apt(state, mirror, distribution, components, enable_backports=False):
    """Setup apt sources and update the cache."""
    logger.info('Setting apt for mirror %s', mirror)
    values = {
        'mirror': mirror,
        'distribution': distribution,
        'components': ' '.join(components)
    }
    basic_template = '''
deb {mirror} {distribution} {components}
deb-src {mirror} {distribution} {components}
'''
    updates_template = '''
deb {mirror} {distribution}-updates {components}
deb-src {mirror} {distribution}-updates {components}
'''
    security_template = '''
deb http://security.debian.org/debian-security/ {distribution}-security {components}
deb-src http://security.debian.org/debian-security/ {distribution}-security {components}
'''
    backports_template = f'''
deb http://deb.debian.org/debian {STABLE_CODENAME}-backports main
deb-src http://deb.debian.org/debian {STABLE_CODENAME}-backports main
'''
    file_path = path_in_mount(state, 'etc/apt/sources.list')
    with open(file_path, 'w') as file_handle:
        file_handle.write(basic_template.format(**values))
        if distribution not in ('sid', 'unstable'):
            file_handle.write(updates_template.format(**values))
            if enable_backports:
                file_handle.write(backports_template)

            file_handle.write(security_template.format(**values))

    run_in_chroot(state, ['apt-get', 'update'])
    run_in_chroot(state, ['apt-get', 'clean'])


def setup_flash_kernel(state, machine_name, kernel_options,
                       boot_filesystem_type):
    """Setup and install flash-kernel package."""
    logger.info('Setting up flash kernel for machine %s with options %s',
                machine_name, kernel_options)
    directory_path = path_in_mount(state, 'etc/flash-kernel')
    os.makedirs(directory_path, exist_ok=True)

    file_path = path_in_mount(state, 'etc/flash-kernel/machine')
    with open(file_path, 'w') as file_handle:
        file_handle.write(machine_name)

    if kernel_options:
        stdin = 'flash-kernel flash-kernel/linux_cmdline string {}'.format(
            kernel_options)
        run_in_chroot(state, ['debconf-set-selections'],
                      feed_stdin=stdin.encode())

    run_in_chroot(state, ['apt-get', 'install', '-y', 'flash-kernel'])

    # flash-kernel creates links in /boot and does not work when the filesystem
    # is vfat.
    if boot_filesystem_type != 'vfat':
        run_in_chroot(state, ['flash-kernel'])


def update_initramfs(state):
    """Update the initramfs in the disk image to make it use fstab etc."""
    logger.info('Updating initramfs')
    run_in_chroot(state, ['update-initramfs', '-u'])


def install_boot_loader_part(state, path, seek, size, count=None):
    """Do a dd copy for a file onto the disk image."""
    image_file = state['image_file']
    full_path = path_in_mount(state, path)
    logger.info('Installing boot loader part %s at seek=%s, size=%s, count=%s',
                full_path, seek, size, count)
    command = [
        'dd', 'if=' + full_path, 'of=' + image_file, 'seek=' + seek,
        'bs=' + size, 'conv=notrunc'
    ]
    if count:
        command.append('count=' + count)

    run(command)


def fill_free_space_with_zeros(state):
    """Fill up the free space in the image with zeros.

    So that we can compress the image better.

    """
    zeros_path = path_in_mount(state, 'ZEROS')
    logger.info('Fill space with zeros on %s', state['mount_point'])
    run(['dd', 'if=/dev/zero', 'of=' + zeros_path, 'bs=1M'], ignore_fail=True)
    run(['rm', '-f', zeros_path])


def compress(archive_file, image_file):
    """Compress an image using xz."""
    logger.info('Compressing file %s to %s', image_file, archive_file)
    command = ['xz', '--no-warn', '--threads=0', '-9', '--force']
    run(command + [image_file])


def sign(archive):
    """Sign an image using GPG."""
    logger.info('Signing file %s with GPG', archive)
    signature = archive + '.sig'
    try:
        os.remove(signature)
    except FileNotFoundError:
        pass

    run(['gpg', '--output', signature, '--detach-sig', archive])



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Base worker class to build all Qemu images.
"""

from .. import library
from .vm import VMImageBuilder


class QemuImageBuilder(VMImageBuilder):
    """Base image builder for all Qemu targets."""
    vm_image_extension = '.qcow2'

    @classmethod
    def get_target_name(cls):
        """Return the name of the target for an image builder."""
        if getattr(cls, 'architecture', None):
            return 'qemu-' + cls.architecture

        return None

    def create_vm_file(self, image_file, vm_file):
        """Create a VM image file from image file."""
        library.run(
            ['qemu-img', 'convert', '-O', 'qcow2', image_file, vm_file])



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Tests for checking Freedom Maker's internal library of actions.
"""

import contextlib
import os
import random
import stat
import string
import tempfile
import unittest
from unittest.mock import Mock, call, patch

from .. import library


class TestLibrary(unittest.TestCase):
    """Test all internal library methods used for building image."""
    def setUp(self):
        """Common setup for each test."""
        self.args = ['1', '2', '3']
        self.kwargs = {'a': 'x', 'b': 'y'}
        self.method = self.random_string()
        self.image = self.random_string()

        self.mount_point_directory = tempfile.TemporaryDirectory()
        os.makedirs(self.mount_point_directory.name + '/usr/sbin/')
        os.makedirs(self.mount_point_directory.name + '/tmp/')
        os.makedirs(self.mount_point_directory.name + '/etc/apt')
        self.state = {
            'image_file': self.image,
            'mount_point': self.mount_point_directory.name,
            'success': True,
        }

    def tearDown(self):
        """Cleanup the test case."""
        self.mount_point_directory.cleanup()

    @staticmethod
    def random_string():
        """Generate a random string."""
        return ''.join(
            [random.choice(string.ascii_lowercase) for _ in range(8)])

    @contextlib.contextmanager
    def assert_file_change(self, path, content, expected_content):
        """Context manager to verify that file contents changed as expected."""
        if content is not None:
            with open(path, 'w') as file_handle:
                file_handle.write(content)

        yield

        with open(path, 'r') as file_handle:
            changed_content = file_handle.read()

        self.assertEqual(expected_content, changed_content)

    @patch('cliapp.runcmd')
    def test_run(self, runcmd):
        """Test the run utility."""
        library.run(*self.args, **self.kwargs)
        assert runcmd.called
        self.assertEqual(list(runcmd.call_args[0]), self.args)
        for key, value in self.kwargs.items():
            self.assertEqual(runcmd.call_args[1][key], value)

        environ = {
            'LC_ALL': 'C',
            'LANGUAGE': 'C',
            'LANG': 'C',
            'DEBIAN_FRONTEND': 'noninteractive',
            'DEBCONF_NONINTERACTIVE_SEEN': 'true'
        }
        for key, value in environ.items():
            runcmd.call_args[1]['env'][key] = value

    @patch('freedommaker.library.run')
    def test_run_in_chroot(self, run):
        """Test executing inside a chroot environment."""
        library.run_in_chroot(self.state, self.args, **self.kwargs)
        expected_args = ['chroot', self.state['mount_point']] + self.args
        self.assertEqual(run.call_args, call(expected_args, **self.kwargs))

    def test_path_in_mount(self):
        """Test returning a sub-directory in mount point."""
        output = library.path_in_mount(self.state, 'boot')
        self.assertEqual(output, self.state['mount_point'] + '/boot')

        output = library.path_in_mount(self.state, '/boot')
        self.assertEqual(output, '/boot')

    def test_schedule_clean(self):
        """Test scheduling cleanup jobs."""
        cleanup = Mock()
        library.schedule_cleanup(self.state, cleanup, 2, b=3)
        self.assertEqual(self.state['cleanup'][-1], [cleanup, (2, ), {'b': 3}])

    def test_cleanup(self):
        """Test the cleanup works."""
        method = Mock()
        library.schedule_cleanup(self.state, method, 1, a=1)
        library.schedule_cleanup(self.state, method, 2, a=2)
        library.schedule_cleanup(self.state, method, 3, a=3)
        library.cleanup(self.state)
        self.assertEqual(method.call_args_list, [((3, ), {
            'a': 3
        }), ((2, ), {
            'a': 2
        }), ((1, ), {
            'a': 1
        })])

    @patch('freedommaker.library.run')
    def test_create_ram_directory_image(self, run):
        """Test that RAM directory is properly created."""
        library.create_ram_directory_image(self.state, self.image, '4G')
        run.assert_called_once_with([
            'mount', '-o', 'size=4G', '-t', 'tmpfs', 'tmpfs',
            self.state['ram_directory'].name
        ])
        self.assertEqual(self.state['image_file'],
                         self.state['ram_directory'].name + '/' + self.image)
        self.assertEqual(
            self.state['cleanup'],
            [[
                library.remove_ram_directory,
                (self.state['ram_directory'], ), {}
            ],
             [
                 library.copy_image,
                 (self.state, self.state['image_file'], self.image), {}
             ]])
        self.state['ram_directory'].cleanup()

    @patch('freedommaker.library.run')
    def test_remove_ram_directory(self, run):
        """Test removing a RAM directory."""
        directory = Mock(name=self.random_string())
        library.remove_ram_directory(directory)
        run.assert_called_once_with(['umount', directory.name])
        directory.cleanup.assert_called()

    @patch('freedommaker.library.run')
    def test_copy_image(self, run):
        """Test copying temp image to final image."""
        temp_image = self.random_string()
        library.copy_image(self.state, temp_image, self.image)
        run.assert_called_once_with(
            ['cp', '--sparse=always', temp_image, self.image])

        run.reset_mock()
        self.state['success'] = False
        library.copy_image(self.state, temp_image, self.image)
        run.assert_called_once_with(
            ['cp', '--sparse=always', temp_image, self.image + '.failed'])

    def test_create_temp_image(self):
        """Test creating a temporary image file on disk."""
        library.create_temp_image(self.state, self.image)
        self.assertEqual(self.state['image_file'], self.image + '.temp')
        self.assertEqual(self.state['cleanup'], [[
            library.move_image,
            (self.state, self.image + '.temp', self.image), {}
        ]])

    @patch('freedommaker.library.run')
    def test_move_image(self, run):
        """Test moving temp image to final image."""
        source_image = self.random_string()
        library.move_image(self.state, source_image, self.image)
        run.assert_called_once_with(['mv', source_image, self.image])

        run.reset_mock()
        self.state['success'] = False
        library.move_image(self.state, source_image, self.image)
        run.assert_called_once_with(
            ['mv', source_image, self.image + '.failed'])

    @patch('freedommaker.library.run')
    def test_create_image(self, run):
        """Test creating an image."""
        library.create_image(self.state, '4G')
        run.assert_called_once_with(
            ['qemu-img', 'create', '-f', 'raw', self.image, '4G'])

    @patch('freedommaker.library.run')
    def test_create_partition_table(self, run):
        """Test creating a partition table."""
        library.create_partition_table(self.state, 'msdos')
        run.assert_called_once_with(
            ['parted', '-s', self.image, 'mklabel', 'msdos'])

    @patch('freedommaker.library.run')
    def test_create_partition(self, run):
        """Test creating a partition table."""
        library.create_partition(self.state, 'root', '10mib', '50%', 'f2fs')
        run.assert_called_once_with([
            'parted', '-s', self.image, 'mkpart', 'primary', 'f2fs', '10mib',
            '50%'
        ])

        self.assertEqual(self.state['partitions'], ['root'])

        library.create_partition(self.state, 'root', '10mib', '50%', 'vfat')
        run.assert_called_with([
            'parted', '-s', self.image, 'mkpart', 'primary', 'fat32', '10mib',
            '50%'
        ])

    @patch('freedommaker.library.run')
    def test_set_boot_flag(self, run):
        """Test that boot flag is properly set."""
        library.set_boot_flag(self.state, 3)
        run.assert_called_with(
            ['parted', '-s', self.image, 'set', '3', 'boot', 'on'])

    @patch('freedommaker.library.run')
    def test_loopback_setup(self, run):
        """Test that loopback device is properly setup."""
        self.state['partitions'] = ['firmware', 'boot', 'root']

        run.return_value = b'''remove x x
add x loop99p1
add x loop99p2
add x loop99p3
modify x x
'''
        library.loopback_setup(self.state)
        run.assert_called_with(['kpartx', '-asv', self.image])
        self.assertEqual(
            self.state['devices'], {
                'firmware': '/dev/mapper/loop99p1',
                'boot': '/dev/mapper/loop99p2',
                'root': '/dev/mapper/loop99p3'
            })
        self.assertEqual(self.state['loop_device'], '/dev/loop99')
        self.assertEqual(
            self.state['cleanup'],
            [[library.force_release_loop_device, ('/dev/loop99', ), {}],
             [
                 library.force_release_partition_loop,
                 ('/dev/mapper/loop99p1', ), {}
             ],
             [
                 library.force_release_partition_loop,
                 ('/dev/mapper/loop99p2', ), {}
             ],
             [
                 library.force_release_partition_loop,
                 ('/dev/mapper/loop99p3', ), {}
             ], [library.loopback_teardown, (self.image, ), {}]])

    @staticmethod
    @patch('freedommaker.library.run')
    def test_force_release_partition_loop(run):
        """Test loop device is forcefully released."""
        library.force_release_partition_loop('/dev/test/loop99')
        run.assert_called_with(['dmsetup', 'remove', '/dev/test/loop99'],
                               ignore_fail=True)

    @staticmethod
    @patch('freedommaker.library.run')
    def test_force_release_loop_device(run):
        """Test loop device is forcefully released."""
        library.force_release_partition_loop('/dev/test/loop99')
        run.assert_called_with(['dmsetup', 'remove', '/dev/test/loop99'],
                               ignore_fail=True)

    @staticmethod
    @patch('freedommaker.library.run')
    def test_loopback_teardown(run):
        """Test tearing down of loopback."""
        library.loopback_teardown('/dev/test/loop99')
        run.assert_called_with(['kpartx', '-dsv', '/dev/test/loop99'])

    @patch('freedommaker.library.run')
    def test_create_filesystem(self, run):
        """Test creating filesystem."""
        library.create_filesystem('/dev/test/loop99p1', 'btrfs')
        self.assertEqual(run.call_args_list, [
            call(['mkfs', '-t', 'btrfs', '/dev/test/loop99p1']),
            call(['udevadm', 'trigger', '/dev/test/loop99p1']),
            call(['udevadm', 'settle'])
        ])

    @patch('freedommaker.library.run')
    def test_mount_filesystem(self, run):
        """Test mounting a filesystem and setting proper state."""
        self.state['devices'] = {
            'root': '/dev/test/loop99p1',
            'firmware': '/dev/test/loop99p2'
        }
        library.mount_filesystem(self.state, 'root', None)
        run.assert_called_with(
            ['mount', '/dev/test/loop99p1', self.state['mount_point']])

        library.mount_filesystem(self.state, 'firmware', 'boot/firmware')
        run.assert_called_with([
            'mount', '/dev/test/loop99p2',
            self.state['mount_point'] + '/boot/firmware'
        ])

        library.mount_filesystem(self.state, '/dev/pts', 'dev/pts',
                                 is_bind_mount=True)
        run.assert_called_with([
            'mount', '/dev/pts', self.state['mount_point'] + '/dev/pts', '-o',
            'bind'
        ])

        sub_mount_points = {
            'root': None,
            'firmware': 'boot/firmware',
            '/dev/pts': 'dev/pts'
        }
        self.assertEqual(self.state['sub_mount_points'], sub_mount_points)

    @patch('freedommaker.library.run')
    def test_unmount_filesystem(self, run):
        """Test unmounting a filesystem."""
        library.unmount_filesystem('/dev/', self.state['mount_point'], False)
        self.assertEqual(
            run.call_args_list,
            [call(['umount', self.state['mount_point']], ignore_fail=False)])

    @patch('freedommaker.library.run')
    def test_process_cleanup(self, run):
        """Test cleaning up processes."""
        library.process_cleanup(self.state)
        self.assertEqual(run.call_args_list, [
            call(['fuser', '-mvk', self.state['mount_point']],
                 ignore_fail=True),
            call(['fuser', '-mvk', self.state['mount_point']],
                 ignore_fail=True)
        ])

    @patch('freedommaker.library.run')
    def test_setup_extra_storage(self, run):
        """Test that setting up extra storage works."""
        extra_storage_file = self.image + '.extra'
        size = '100M'
        mount_point = self.state['mount_point']
        loop_device = self.random_string()
        run.return_value = f'{loop_device}\n'.encode()

        library.setup_extra_storage(self.state, 'ext4', '100M')
        run.assert_not_called()

        library.setup_extra_storage(self.state, 'btrfs', '100M')
        run.assert_has_calls([
            call(['qemu-img', 'create', '-f', 'raw', extra_storage_file,
                  size]),
            call(['losetup', '--show', '--find', extra_storage_file]),
            call(['btrfs', 'device', 'add', loop_device, mount_point])
        ])

        self.assertEqual(self.state['cleanup'], [[
            library.cleanup_extra_storage,
            (self.state, loop_device, extra_storage_file), {}
        ]])

    @patch('freedommaker.library.run')
    def test_cleanup_extra_storage(self, run):
        """Test that extra storage will be cleaned properly."""
        extra_storage_file = self.image + '.extra'
        loop_device = self.random_string()
        mount_point = self.state['mount_point']

        library.cleanup_extra_storage(self.state, loop_device,
                                      extra_storage_file)

        run.assert_has_calls([
            call(['btrfs', 'balance', 'start', '-musage=0', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-dusage=0', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-musage=20', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-dusage=20', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-musage=40', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-dusage=40', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-musage=60', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-dusage=60', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-musage=80', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-dusage=80', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-mconvert=dup', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'device', 'remove', loop_device, mount_point]),
            call(['losetup', '--detach', loop_device]),
            call(['rm', '-f', extra_storage_file]),
            call(['btrfs', 'balance', 'start', '-musage=0', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-dusage=0', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-musage=20', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-dusage=20', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-musage=40', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-dusage=40', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-musage=60', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-dusage=60', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-musage=80', mount_point],
                 ignore_fail=True),
            call(['btrfs', 'balance', 'start', '-dusage=80', mount_point],
                 ignore_fail=True),
        ])

    @patch('freedommaker.library.run')
    def test_debootstrap(self, run):
        """Test debootstrapping."""
        library.debootstrap(self.state, 'i386', 'stretch', 'minbase',
                            ['main', 'contrib'], ['p1', 'p2'],
                            'http://deb.debian.org/debian')
        run.assert_called_with([
            'debootstrap', '--arch=i386', '--variant=minbase',
            '--components=main,contrib', '--include=p1,p2', 'stretch',
            self.state['mount_point'], 'http://deb.debian.org/debian'
        ])

        self.assertEqual(self.state['cleanup'], [[
            library.unmount_filesystem,
            (None, self.state['mount_point'] + '/etc/machine-id'), {
                'ignore_fail': True
            }
        ]])

    def test_no_daemon_policy(self):
        """Test that no daemon run policy is properly set."""
        file_path = self.state['mount_point'] + '/usr/sbin/policy-rc.d'
        with library.no_run_daemon_policy(self.state):
            with open(file_path, 'r') as file_handle:
                contents = file_handle.read()

            self.assertEqual(contents, '#!/bin/sh\nexit 101\n')
            self.assertEqual(oct(os.stat(file_path)[stat.ST_MODE])[-3:], '755')

        self.assertFalse(os.path.isfile(file_path))

    @patch('freedommaker.library.run_in_chroot')
    def test_install_package(self, run):
        """Test installing a package."""
        library.install_package(self.state, 'nmap')
        run.assert_called_with(self.state,
                               ['apt-get', 'install', '-y', 'nmap'])
        run.reset_mock()

    @patch('freedommaker.library.run_in_chroot')
    def test_install_from_backports(self, run):
        library.install_from_backports(
            self.state,
            ('freedombox', 'freedombox-doc-en', 'freedombox-doc-es'))

        self.assertEqual(run.call_args_list, [
            call(self.state, [
                'apt-get', 'install', '-y',
                f'freedombox/{library.STABLE_CODENAME}-backports',
                f'freedombox-doc-en/{library.STABLE_CODENAME}-backports',
                f'freedombox-doc-es/{library.STABLE_CODENAME}-backports'
            ]),
            call(self.state, ['apt-get', 'autoremove', '-y'])
        ])

    @patch('freedommaker.library.install_package')
    @patch('freedommaker.library.run_in_chroot')
    def test_install_custom_package(self, run, install_package):
        """Test installing a custom package."""
        with tempfile.NamedTemporaryFile() as file_path:
            library.install_custom_package(self.state, file_path.name)

            run.assert_called_with(self.state, [
                'apt', 'install', '-y', '--allow-downgrades',
                '/tmp/' + os.path.basename(file_path.name)
            ])

    def test_set_hostname(self):
        """Test that hostname is properly set."""
        hosts_path = self.state['mount_point'] + '/etc/hosts'
        hostname_path = self.state['mount_point'] + '/etc/hostname'

        content = '''127.0.0.1 localhost
::1     localhost ip6-localhost ip6-loopback
ff02::1 ip6-allnodes
ff02::2 ip6-allrouters
'''
        expected_content = content + '''127.0.1.1 fbx\n'''
        with self.assert_file_change(hosts_path, content, expected_content):
            with self.assert_file_change(hostname_path, 'test', 'fbx\n'):
                library.set_hostname(self.state, 'fbx')

    def test_get_fstab_options(self):
        """Test getting fstab options for a file system."""
        options = library.get_fstab_options('ext4')
        self.assertEqual(options, 'errors=remount-ro')

        options = library.get_fstab_options('btrfs')
        self.assertEqual(options, 'compress=zstd:1')

    @patch('freedommaker.library.run', return_value=b'test-uuid')
    def test_get_uuid_of_device(self, run):
        """Test retrieving UUID of device."""
        response = library.get_uuid_of_device('/dev/test/loop99p1')
        run.assert_called_with([
            'blkid', '--output=value', '--match-tag=UUID', '/dev/test/loop99p1'
        ])
        self.assertEqual(response, 'test-uuid')

    @patch('freedommaker.library.get_uuid_of_device')
    def test_add_fstab_entry(self, get_uuid):
        """Test adding entries to /etc/fstab."""
        fstab_path = self.state['mount_point'] + '/etc/fstab'
        self.state['devices'] = {
            'root': '/dev/test/loop99p1',
            'boot': '/dev/test/loop99p2'
        }
        self.state['sub_mount_points'] = {'root': None, 'boot': 'boot'}

        expected_content = 'UUID=root-uuid / btrfs compress=zstd:1 0 1\n'
        get_uuid.return_value = 'root-uuid'
        with self.assert_file_change(fstab_path, 'initial-trash',
                                     expected_content):
            library.add_fstab_entry(self.state, 'root', 'btrfs', 1,
                                    append=False)

        expected_content += 'UUID=boot-uuid /boot ext4 errors=remount-ro 0 2\n'
        get_uuid.return_value = 'boot-uuid'
        with self.assert_file_change(fstab_path, None, expected_content):
            library.add_fstab_entry(self.state, 'boot', 'ext4', 2, append=True)

    @patch('freedommaker.library.run_in_chroot')
    def test_install_grub(self, run):
        """Test installing grub boot loader."""
        self.state['loop_device'] = '/dev/test/loop99'
        library.install_grub(self.state)
        self.assertEqual(run.call_args_list, [
            call(self.state, ['update-grub']),
            call(self.state, ['grub-install', '/dev/test/loop99'])
        ])

        run.reset_mock()
        library.install_grub(self.state, target='arm-efi')
        self.assertEqual(run.call_args_list, [
            call(self.state, ['update-grub']),
            call(self.state,
                 ['grub-install', '/dev/test/loop99', '--target=arm-efi'])
        ])

        run.reset_mock()
        library.install_grub(self.state, target='arm-efi', is_efi=True)
        self.assertEqual(run.call_args_list, [
            call(self.state, ['update-grub']),
            call(self.state, [
                'grub-install', '/dev/test/loop99', '--target=arm-efi',
                '--no-nvram'
            ])
        ])

    @patch('freedommaker.library.run_in_chroot')
    def test_setup_apt(self, run):
        """Test setting up apt."""
        sources_path = self.state['mount_point'] + '/etc/apt/sources.list'

        stable_content = '''
deb http://deb.debian.org/debian stable main
deb-src http://deb.debian.org/debian stable main

deb http://deb.debian.org/debian stable-updates main
deb-src http://deb.debian.org/debian stable-updates main

deb http://security.debian.org/debian-security/ stable-security main
deb-src http://security.debian.org/debian-security/ stable-security main
'''
        with self.assert_file_change(sources_path, None, stable_content):
            library.setup_apt(self.state, 'http://deb.debian.org/debian',
                              'stable', ['main'])

        self.assertEqual(run.call_args_list, [
            call(self.state, ['apt-get', 'update']),
            call(self.state, ['apt-get', 'clean'])
        ])

        unstable_content = '''
deb http://ftp.us.debian.org/debian unstable main contrib non-free
deb-src http://ftp.us.debian.org/debian unstable main contrib non-free
'''
        with self.assert_file_change(sources_path, None, unstable_content):
            library.setup_apt(self.state, 'http://ftp.us.debian.org/debian',
                              'unstable', ['main', 'contrib', 'non-free'])

    @patch('freedommaker.library.run_in_chroot')
    def test_setup_flash_kernel(self, run):
        """Test setting up flash kernel."""
        machine_path = self.state['mount_point'] + '/etc/flash-kernel/machine'
        expected_content = 'test-machine'
        with self.assert_file_change(machine_path, None, expected_content):
            library.setup_flash_kernel(self.state, 'test-machine', None,
                                       'ext2')

        self.assertEqual(run.call_args_list, [
            call(self.state, ['apt-get', 'install', '-y', 'flash-kernel']),
            call(self.state, ['flash-kernel'])
        ])

        run.reset_mock()
        with self.assert_file_change(machine_path, None, expected_content):
            library.setup_flash_kernel(self.state, 'test-machine', 'debug',
                                       'vfat')

        selection = b'flash-kernel flash-kernel/linux_cmdline string debug'
        self.assertEqual(run.call_args_list, [
            call(self.state, ['debconf-set-selections'], feed_stdin=selection),
            call(self.state, ['apt-get', 'install', '-y', 'flash-kernel']),
        ])

    @patch('freedommaker.library.run_in_chroot')
    def test_update_initramfs(self, run):
        """Test updating initramfs."""
        library.update_initramfs(self.state)

        self.assertEqual(run.call_args_list,
                         [call(self.state, ['update-initramfs', '-u'])])

    @patch('freedommaker.library.run')
    def test_install_boot_loader_path(self, run):
        """Test installing boot loader components using dd."""
        path = 'u-boot/path'
        full_path = self.state['mount_point'] + '/' + path
        library.install_boot_loader_part(self.state, path, '533', '515')
        run.assert_called_with([
            'dd', 'if=' + full_path, 'of=' + self.image, 'seek=533', 'bs=515',
            'conv=notrunc'
        ])

        library.install_boot_loader_part(self.state, path, '533', '515', '90')
        run.assert_called_with([
            'dd', 'if=' + full_path, 'of=' + self.image, 'seek=533', 'bs=515',
            'conv=notrunc', 'count=90'
        ])

    @patch('freedommaker.library.run')
    def test_fill_free_space_with_zeros(self, run):
        """Test filling free space with zeros."""
        zeros_path = self.state['mount_point'] + '/ZEROS'
        library.fill_free_space_with_zeros(self.state)
        self.assertEqual(run.call_args_list, [
            call(['dd', 'if=/dev/zero', 'of=' + zeros_path, 'bs=1M'],
                 ignore_fail=True),
            call(['rm', '-f', zeros_path])
        ])

    @patch('freedommaker.library.run')
    def test_compress(self, run):
        """Test compressing an image."""
        archive_file = self.random_string()
        image_file = self.random_string()

        library.compress(archive_file, image_file)
        run.assert_called_with(
            ['xz', '--no-warn', '--threads=0', '-9', '--force', image_file])

    @patch('os.remove')
    @patch('freedommaker.library.run')
    def test_sign(self, run, remove):
        """Test running signing with GPG."""
        archive = self.random_string()
        signature = archive + '.sig'
        remove.side_effect = FileNotFoundError
        library.sign(archive)
        run.assert_called_with(
            ['gpg', '--output', signature, '--detach-sig', archive])



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build AMD64 images.
"""

from .amd_intel import AMDIntelImageBuilder


class AMD64ImageBuilder(AMDIntelImageBuilder):
    """Image builder for all amd64 targets."""
    architecture = 'amd64'
    kernel_flavor = 'amd64'



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build Pine64+ image.
"""

from .a64 import A64ImageBuilder


class Pine64PlusImageBuilder(A64ImageBuilder):
    """Image builder for Pine64+ target."""
    machine = 'pine64-plus'
    flash_kernel_name = 'Pine64+'
    u_boot_target = 'pine64_plus'



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build a universal ARM64/ARM32 images using UEFI.
"""

import pathlib

from .. import library
from ..builder import ImageBuilder

DTB_SCRIPT = '''#!/bin/sh

set -e

if [ ! -e /boot/efi/ ]; then
    exit 0
fi

# See Chapter 8 of Debian Linux Kernel Handbook
set -- $DEB_MAINT_PARAMS
action="$1"
action="${action#\'}"
action="${action%\'}"

# Only call once on install, upgrade, removal or purge
hook="$(basename "$(dirname "$0")")"
case "$hook:$action" in
  postinst.d:configure|postinst.d:|postrm.d:remove|postrm.d:)
    latest_version=$(linux-version list | linux-version sort | tail -1)
    source_dir="/usr/lib/linux-image-${latest_version}"
    if [ -d "${source_dir}" ]; then
      echo "Copying DTBs from ${source_dir} to EFI partition" >&2
      rm -rf /boot/efi/dtb
      cp -r "${source_dir}" /boot/efi/dtb
    fi
  ;;
esac
'''


class ARMEFIImageBuilder(ImageBuilder):
    """Image builder a universal ARM64/ARM32 images using UEFI."""
    efi_architecture = None
    partition_table_type = 'gpt'
    efi_filesystem_type = 'vfat'
    efi_size = '256MiB'
    boot_loader = None
    grub_target = None

    def __init__(self, *args, **kwargs):
        """Initialize the object."""
        super().__init__(*args, **kwargs)
        if not self.efi_architecture:
            raise Exception('EFI architecture is not specified')

    @classmethod
    def install_boot_loader(cls, state):
        """Install the DTB files and boot manager."""
        library.install_grub(state, target=cls.grub_target, is_efi=True)

        # In typical UEFI systems, EFI boot manager is available and all the OS
        # list themselves as boot manager configuration in NVRAM. In many SBCs,
        # NVRAM may not be present. In our case, since we distribute disk
        # images, we don't have other operating system to boot. So, install
        # grub as the boot manager.
        library.run_in_chroot(state, ['mkdir', '-p', '/boot/efi/EFI/boot/'])
        library.run_in_chroot(state, [
            'cp', f'/boot/efi/EFI/debian/grub{cls.efi_architecture}.efi',
            f'/boot/efi/EFI/boot/boot{cls.efi_architecture}.efi'
        ])

        # Grub does not have a way of providing the kernel with an appropriate
        # DTB file need to understand the hardware layout of an SBC. u-boot
        # firmware provided by the SBC manufacturer can perform this task if
        # the DTB suitable for the board is available in /dtb/ directory of the
        # EFI partition.
        usr_lib = pathlib.Path(library.path_in_mount(state, 'usr/lib/'))
        linux = list(usr_lib.glob('linux-image-*'))[0].name
        library.run_in_chroot(
            state, ['cp', '-r', f'/usr/lib/{linux}/', '/boot/efi/dtb'])

        # Install a permanent script that would continuously update DTBs after
        # newer kernels are installed.
        dir1 = library.path_in_mount(state, 'etc/kernel/postinst.d/')
        dir2 = library.path_in_mount(state, 'etc/kernel/postrm.d/')
        for directory in [dir1, dir2]:
            script = pathlib.Path(directory) / 'zz-freedombox-efi-dtbs'
            script.write_text(DTB_SCRIPT)
            script.chmod(0o755)



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build a universal ARM64 images using UEFI.
"""

from .arm_efi import ARMEFIImageBuilder


class ARM64ImageBuilder(ARMEFIImageBuilder):
    """Image builder a universal ARM64 images using UEFI."""
    architecture = 'arm64'
    kernel_flavor = 'arm64'
    efi_architecture = 'aa64'

    def __init__(self, *args, **kwargs):
        """Initialize builder object."""
        super().__init__(*args, **kwargs)
        self.packages += ['grub-efi-arm64']

    @classmethod
    def get_target_name(cls):
        """Return the command line name of target for this builder."""
        return 'arm64'



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build VirtualBox i386 images.
"""

from .virtualbox import VirtualBoxImageBuilder


class VirtualBoxI386ImageBuilder(VirtualBoxImageBuilder):
    """Image builder for all VirtualBox i386 targets."""
    architecture = 'i386'
    kernel_flavor = '686'



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build TI AM335x BeagleBone Black image.
"""

from .. import library
from .arm import ARMImageBuilder


class BeagleBoneImageBuilder(ARMImageBuilder):
    """Image builder for BeagleBone target."""
    architecture = 'armhf'
    machine = 'beaglebone'
    kernel_flavor = 'armmp'
    flash_kernel_name = 'TI AM335x BeagleBone Black'
    flash_kernel_options = 'console=ttyO0'

    @staticmethod
    def install_boot_loader(state):
        """Install the boot loader onto the image."""
        library.install_boot_loader_part(state,
                                         'usr/lib/u-boot/am335x_boneblack/MLO',
                                         seek='1',
                                         size='128k',
                                         count='1')
        library.install_boot_loader_part(
            state,
            'usr/lib/u-boot/am335x_boneblack/u-boot.img',
            seek='1',
            size='384k',
            count='2')



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build Qemu AMD64 images.
"""

from .qemu import QemuImageBuilder


class QemuAmd64ImageBuilder(QemuImageBuilder):
    """Image builder for all Qemu amd64 targets."""
    architecture = 'amd64'
    kernel_flavor = 'amd64'



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build a universal ARM32 images using UEFI.
"""

from .arm_efi import ARMEFIImageBuilder


class ARMHFImageBuilder(ARMEFIImageBuilder):
    """Image builder a universal ARM32 images using UEFI."""
    architecture = 'armhf'
    kernel_flavor = 'armmp-lpae'
    efi_architecture = 'arm'
    grub_target = 'arm-efi'

    def __init__(self, *args, **kwargs):
        """Initialize builder object."""
        super().__init__(*args, **kwargs)
        self.packages += ['grub-efi-arm']

    @classmethod
    def get_target_name(cls):
        """Return the command line name of target for this builder."""
        return 'armhf'



#!/usr/bin/python3
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Module to provide ability to run this freedommaker package.

Run the package as:
  python3 -m freedommaker
"""

from .application import Application

if __name__ == '__main__':
    Application().run()



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build Vagrant images.
"""

import os

from .. import library
from .virtualbox_amd64 import VirtualBoxAmd64ImageBuilder


class VagrantImageBuilder(VirtualBoxAmd64ImageBuilder):
    """Image builder for Vagrant package."""
    vagrant_extension = '.box'

    @classmethod
    def get_target_name(cls):
        """Return the name of the target for an image builder."""
        return 'vagrant'

    def build(self):
        """Run the image building process."""
        vm_file = self._replace_extension(self.image_file,
                                          self.vm_image_extension)
        vagrant_file = self._replace_extension(self.image_file,
                                               self.vagrant_extension)

        self.make_image()
        self.create_vm_file(self.image_file, vm_file)
        os.remove(self.image_file)
        self.vagrant_package(vm_file, vagrant_file)

    def vagrant_package(self, vm_file, vagrant_file):
        """Create a vagrant package from VM file."""
        command = [
            'bin/vagrant-package', '--distribution',
            self.arguments.distribution, '--release-components'
        ]
        command.extend(self.release_components)
        command += ['--output', vagrant_file, vm_file]
        library.run(command)



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build Lamobo R1 (BananaPI Router v1) image.
"""

from .a20 import A20ImageBuilder


class LamoboR1ImageBuilder(A20ImageBuilder):
    """Image builder for Lamobo R1 (BananaPI Router v1) targets."""
    machine = 'lamobo-r1'
    flash_kernel_name = 'Lamobo R1'
    u_boot_path = 'usr/lib/u-boot/Lamobo_R1/u-boot-sunxi-with-spl.bin'



#!/usr/bin/python3
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Command line application wrapper over image builder.
"""

import argparse
import datetime
import logging
import logging.config
import os

import freedommaker

from .builder import ImageBuilder

IMAGE_SIZE = '3800M'
BUILD_MIRROR = 'http://deb.debian.org/debian'
MIRROR = 'http://deb.debian.org/debian'
DISTRIBUTION = 'unstable'
BUILD_DIR = 'build'
LOG_LEVEL = 'debug'
HOSTNAME = 'freedombox'

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Application(object):
    """Command line application to build FreedomBox images."""
    def __init__(self):
        """Initialize object."""
        self.arguments = None

    def run(self):
        """Parse the command line args and execute the command."""
        self.parse_arguments()

        self.setup_logging()
        logger.info('Freedom Maker version - %s', freedommaker.__version__)

        try:
            logger.info('Creating directory - %s', self.arguments.build_dir)
            os.makedirs(self.arguments.build_dir)
        except os.error:
            pass

        for target in self.arguments.targets:
            logger.info('Building target - %s', target)

            cls = ImageBuilder.get_builder_class(target)
            if not cls:
                logger.warning('Unknown target - %s', target)
                continue

            builder = cls(self.arguments)
            try:
                builder.build()
                logger.info('Target complete - %s', target)
            except:  # noqa: E722
                logger.error('Target failed - %s', target)
                raise

    def parse_arguments(self):
        """Parse command line arguments."""
        build_stamp = datetime.datetime.today().strftime('%Y-%m-%d')

        parser = argparse.ArgumentParser(
            description='FreedomMaker - Script to build FreedomBox images',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser.add_argument('--build-stamp', default=build_stamp,
                            help='Build stamp to use on image file names')
        parser.add_argument('--image-size', default=IMAGE_SIZE,
                            help='Size of the image to build')
        parser.add_argument('--build-mirror', default=BUILD_MIRROR,
                            help='Debian mirror to use for building')
        parser.add_argument('--mirror', default=MIRROR,
                            help='Debian mirror to use in built image')
        parser.add_argument('--distribution', default=DISTRIBUTION,
                            help='Debian release to use in built image')
        parser.add_argument('--add-release-component', action='append',
                            dest='release_component',
                            help='Add an extra Debian release component '
                                 '(other than main)')
        parser.add_argument('--package', action='append',
                            help='Install additional packages in the image')
        parser.add_argument(
            '--custom-package', action='append',
            help='Install package from DEB file into the image')
        parser.add_argument(
            '--enable-backports', action='store_true',
            help='Deprecated: Backports are now enabled for '
            'stable images by default')
        parser.add_argument('--disable-backports', action='store_true',
                            help='Disable backports in the image')
        parser.add_argument(
            '--build-dir', default=BUILD_DIR,
            help='Directory to build images and create log file')
        parser.add_argument(
            '--log-level', default=LOG_LEVEL, help='Log level',
            choices=('critical', 'error', 'warn', 'info', 'debug'))
        parser.add_argument('--hostname', default=HOSTNAME,
                            help='Hostname to set inside the built images')
        parser.add_argument(
            '--sign', action='store_true',
            help='Sign the images with default GPG key after building')
        parser.add_argument(
            '--force', action='store_true',
            help='Force rebuild of images even when required image exists')
        parser.add_argument(
            '--build-in-ram', action='store_true',
            help='Build the image in RAM so that it is faster, requires '
            'free RAM about the size of disk image')
        parser.add_argument('--skip-compression', action='store_true',
                            help='Do not compress the generated image')
        parser.add_argument('--with-build-dep', action='store_true',
                            help='Include build dependencies in the image')
        parser.add_argument('targets', nargs='+',
                            help='Image targets to build')

        self.arguments = parser.parse_args()

    def setup_logging(self):
        """Setup logging."""
        config = {
            'version': 1,
            'formatters': {
                'date': {
                    'format': '%(asctime)s - %(levelname)s - %(message)s'
                }
            },
            'handlers': {
                'console': {
                    'class': 'logging.StreamHandler',
                    'formatter': 'date',
                },
            },
            'root': {
                'level': self.arguments.log_level.upper(),
                'handlers': ['console'],
            },
            'disable_existing_loggers': False
        }
        logging.config.dictConfig(config)



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build Pine64 LTS image.
"""

from .a64 import A64ImageBuilder


class Pine64LTSImageBuilder(A64ImageBuilder):
    """Image builder for Pine64 LTS target."""
    machine = 'pine64-lts'
    flash_kernel_name = 'Pine64 LTS'
    u_boot_target = 'pine64-lts'



#!/usr/bin/python3
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Freedom Maker setup file
"""

import setuptools

from freedommaker import __version__

setuptools.setup(
    name='freedom-maker',
    version=__version__,
    description='The FreedomBox image builder',
    author='FreedomBox Authors',
    author_email='freedombox-discuss@lists.alioth.debian.org',
    url='http://freedomboxfoundation.org',
    packages=setuptools.find_packages(),
    scripts=['bin/passwd-in-image', 'bin/vagrant-package'],
    entry_points={'console_scripts': ['freedom-maker = freedommaker:main']},
    test_suite='freedommaker.tests',
    license='COPYING',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: DFSG approved',
        'License :: OSI Approved :: '
        'GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Unix Shell',
        'Topic :: System :: Software Distribution',
    ],
    include_package_data=True,
)



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Base worker class to build Virtual Machine images.
"""

import os

from .amd_intel import AMDIntelImageBuilder


class VMImageBuilder(AMDIntelImageBuilder):
    """Base image builder for all virtual machine targets."""
    vm_image_extension = None

    def build(self):
        """Run the image building process."""
        vm_file = self._replace_extension(self.image_file,
                                          self.vm_image_extension)
        vm_archive_file = vm_file + '.xz'

        self.make_image()
        self.create_vm_file(self.image_file, vm_file)
        os.remove(self.image_file)
        self.compress(vm_archive_file, vm_file)

        self.sign(vm_archive_file)

    def create_vm_file(self, image_file, vm_file):
        """Create a VM image from image file."""
        raise Exception('Not reached')



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Tests for checking built image with VirtualBox.
"""

import logging
import os
import random
import string
import subprocess
import time
import unittest

logger = logging.getLogger(__name__)


class TestVirtualBox(unittest.TestCase):
    """Tests for checking built image with VirtualBox."""
    def setUp(self):
        """Setup test case."""
        self.path = os.path.dirname(__file__)
        self.binary = 'freedommaker'
        self.output_dir = os.path.join(self.path, 'output')

        self.build_stamp = self.random_string()
        self.current_target = 'virtualbox-i386'

    def random_string(self):
        """Generate a random string."""
        return ''.join(
            [random.choice(string.ascii_lowercase) for _ in range(8)])

    def invoke(self, targets=None, **kwargs):
        """Invoke Freedom Maker."""
        parameters = ['--build-dir', self.output_dir]

        if 'build_stamp' not in kwargs:
            parameters += ['--build-stamp', self.build_stamp]

        command = ['python3', '-m', self.binary] + parameters + targets
        subprocess.check_call(command)

    def get_built_file(self):
        """Return the path of the expected built file."""
        target = self.current_target
        extra_map = {
            'virtualbox-amd64': 'all-amd64.vdi',
            'virtualbox-i386': 'all-i386.vdi',
        }

        file_name = 'freedombox-unstable-free_{build_stamp}_{extra}.xz' \
            .format(build_stamp=self.build_stamp, extra=extra_map[target])

        return os.path.join(self.output_dir, file_name)

    def _run(self, *args, ignore_errors=False):
        """Execute a command."""
        subprocess.run(*args, check=not ignore_errors)

    @unittest.skipUnless(
        os.environ.get('FM_RUN_VM_TESTS') == 'true', 'Not requested')
    def test_basic_build(self):
        """Test booting and opening SSH shell.

        Also:
          - Output the plinth diagnostic log.
        """
        self.invoke(['virtualbox-i386'])

        vm_name = 'freedom-maker-test'
        test_ssh_port = 2222
        first_run_wait_time = 120

        compressed_built_file = self.get_built_file()
        built_file = compressed_built_file.rsplit('.', maxsplit=1)[0]

        if not os.path.isfile(built_file):
            subprocess.check_call(
                ['unxz', '--keep', '--force', compressed_built_file])

        passwd_tool = os.path.join(self.path, '..', '..', 'bin',
                                   'passwd-in-image')
        self._run([
            'sudo', 'python3', passwd_tool, built_file, 'fbx', '--password',
            'frdm'
        ])
        try:
            self._run([
                'VBoxManage', 'createvm', '--name', vm_name, '--ostype',
                'Debian', '--register'
            ])
            self._run([
                'VBoxManage', 'storagectl', vm_name, '--name',
                'SATA Controller', '--add', 'sata', '--controller', 'IntelAHCI'
            ])
            self._run([
                'VBoxManage', 'storageattach', vm_name, '--storagectl',
                'SATA Controller', '--port', '0', '--device', '0', '--type',
                'hdd', '--medium', built_file
            ])
            self._run([
                'VBoxManage', 'modifyvm', vm_name, '--pae', 'on', '--memory',
                '1024', '--vram', '128', '--nic1', 'nat', '--natpf1',
                ',tcp,,{port},,22'.format(port=test_ssh_port)
            ])
            self._run(['VBoxManage', 'startvm', vm_name, '--type', 'headless'])
            time.sleep(first_run_wait_time)

            echo = subprocess.Popen(['echo', 'frdm'], stdout=subprocess.PIPE)
            process = subprocess.Popen([
                'sshpass', '-p', 'frdm', 'ssh', '-o',
                'UserKnownHostsFile=/dev/null', '-o',
                'StrictHostKeyChecking=no', '-t', '-t', '-p',
                str(test_ssh_port), 'fbx@127.0.0.1',
                'sudo plinth --list-modules'
            ],
                                       stdin=echo.stdout)
            process.communicate()
        finally:
            echo = subprocess.Popen(['echo', 'frdm'], stdout=subprocess.PIPE)
            process = subprocess.Popen([
                'sshpass', '-p', 'frdm', 'ssh', '-o',
                'UserKnownHostsFile=/dev/null', '-o',
                'StrictHostKeyChecking=no', '-t', '-t', '-p',
                str(test_ssh_port), 'fbx@127.0.0.1', 'sudo shutdown now'
            ],
                                       stdin=echo.stdout)
            process.communicate()
            time.sleep(30)
            self._run(['VBoxManage', 'modifyvm', vm_name, '--hda', 'none'],
                      ignore_errors=True)
            self._run(['VBoxManage', 'unregistervm', vm_name, '--delete'],
                      ignore_errors=True)



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Worker class to build i386 images.
"""

from .amd_intel import AMDIntelImageBuilder


class I386ImageBuilder(AMDIntelImageBuilder):
    """Image builder for all i386 targets."""
    architecture = 'i386'
    kernel_flavor = '686'
